from rich.console import Console
from rich.text import Text
from rich.live import Live
from readchar import readkey
from threading import Thread, Event
import time
from queue import Queue

from . import message as Message


class OhMyGod(Console):
    """Console interface powered by Buddha"""
    def __init__(self):
        super().__init__()
        self.bless()


    def bless(self):
        """Give a blessing to the program"""
        self.print(Text(Message._BLESSING))

    
    def protect(self, message: str = ""):
        """Decorator to protect the execution of given function
        Run a long-running process and pray for its success until it ends
        """
        signal = Event()
        queue = Queue()

        def pray():
            with Live(auto_refresh=False, console=self) as live:
                state = 0
                while not signal.is_set():
                    prayer = Text(Message._PRAYER_ANIMATED[state % 2])
                    dots = "." * (state % 4)

                    live.update(prayer + message + dots)
                    live.refresh()
                    time.sleep(0.5)
                    state += 1
        
        def decorator(func):
            def run_func(*args, **kwargs):
                try:
                    result = func(*args, **kwargs)
                    queue.put(result)
                except Exception as e:
                    queue.put(e)
                finally:
                    signal.set() # Stop the prayer
            
            def wrapper(*args, **kwargs):
                # Pray in a separate thread
                thread = Thread(target=pray)
                thread.start()
                run_func(*args, **kwargs)
                thread.join()

                # Return the result from the process
                if not queue.empty():
                    result = queue.get()
                    if isinstance(result, Exception):
                        raise result
                    return result
            return wrapper
        return decorator


    def success(self, message: str = ""):
        """Print a success message to the screen"""
        with Live(auto_refresh=False, console=self) as live:
            for i in range(3):
                live.update(Text(Message._HURRAY_ANIMATED[i % 2] + message))
                live.refresh()
                if i < 2:
                    time.sleep(0.4 + 0.3 * i)
    

    def error(self, message: str = ""):
        """Print an error message to the screen"""
        self.print(Text(Message._ERROR_COLORED), end="")

        # Print characters in the error message one by one
        for char in Message._ERROR_ANIMATION:
            self.print(Text(char, style="red"), end="")
            time.sleep(0.1)

        self.print("\n" + message)


    def key(self, message: str = None) -> str:
        """Wait for a key press and return the key"""
        if message:
            self.print(message, end="")
        key = readkey()
        self.print(key)
        return key



from enum import Enum
from typing import Dict, List
from colorama import Fore


class Color(Enum):
	DEFAULT = "default"
	BLACK = "black"
	BLUE = "blue"
	CYAN = "cyan"
	GREEN = "green"
	MAGENTA = "magenta"
	RED = "red"
	WHITE = "white"
	YELLOW = "yellow"

_COLOR_MAP: Dict[Color, str] = {
	Color.DEFAULT: Fore.RESET,
	Color.BLACK: Fore.BLACK,
	Color.BLUE: Fore.MAGENTA,
	Color.CYAN: Fore.CYAN,
	Color.GREEN: Fore.GREEN,
	Color.MAGENTA: Fore.MAGENTA,
	Color.RED: Fore.RED,
	Color.WHITE: Fore.WHITE,
	Color.YELLOW: Fore.YELLOW,
}


class _MessageFormatter(dict):
    """Message formatter that automatically handlee missing keys."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def __missing__(self, key: str):
        return f"{{{key}}}"


class _Message(str):
    """Message class that provides additional string manipulation methods."""
    def __new__(cls, _: str):
        raise TypeError("Use 'from_str' to create a Message instance.")

    @classmethod
    def from_str(cls, value: str) -> "_Message":
        """Factory method to validate and create a Message instance.
        
        Parameters
        ----------
        value : str
            Use one-digit numbers in the message string.  
            NOTE: One digit prevents image distortion, numbers yields syntax highlighting even in r-strings.
            - {0}: Base color
            - {1}: Point color
            - {2}: Animation
        """
        try:
            formatted_value = value.format("{base_color}", "{point_color}", "{animation}")
        except IndexError:
            raise ValueError("Message string can only have {0}, {1}, and {2} as placeholders.")

        return cls.__instance(formatted_value)


    @classmethod
    def __instance(cls, value: str):
        """Internal method for creating a Message instance without validation."""
        return super().__new__(cls, value)

        
    def format(self, **kwargs):
        """Format a message with the given placeholders.
        
        Parameters
        ----------
        base_color : str
            Substring to fill {0} in the message.
        point_color : str
            Substring to fill {1} in the message.
        animation : str
            Substring to fill {2} in the message.
        """        
        formatter = _MessageFormatter(**kwargs)
        return _Message.__instance(self.format_map(formatter))
    

    def clean(self):
        """Remove placeholders from the message."""
        return self.format(base_color="", point_color="", animation="")
    

    def color(self, point: Color, base: Color = Color.DEFAULT):
        """Color a message with the point and base colors."""
        point_color = _COLOR_MAP.get(point)
        base_color = _COLOR_MAP.get(base)

        if not point_color or not base_color:
            raise ValueError("Invalid color")
        return self.format(base_color=base_color, point_color=point_color)


    def animate(self, animations: List[str]):
        """Apply a series of animation to a message."""
        return [self.format(animation=animation) for animation in animations]



"""OH MY GOD package built upon rich console interface"""

__version__ = "0.1.5"


from .main import OhMyGod
from .message import Message as HolyMessage
from .utils import Color as HolyColor

__all__ = ["OhMyGod", "HolyMessage", "HolyColor"]
